
   /*
    * little support functions 
    *
    */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include "misc.h"

#undef DEBUG

extern int verbose;


   /*********************************************************************
    * ROUTINE: shMalloc
    *
    * Attempt to allocate the given number of bytes.  Return the 
    * memory, if we succeeded, or print an error message and
    * exit with error code if we failed.
    * 
    * RETURNS:
    *      void *             to new memory, if we got it
    */

void *
shMalloc
   (
   int nbytes                /* I: allocate a chunk of this many bytes */
   )
{
   void *vptr;

   if ((vptr = (void *) malloc(nbytes)) == NULL) {
      shError("shMalloc: failed to allocate for %d bytes", nbytes);
      exit(1);
   }
   return(vptr);
}


   /*********************************************************************
    * ROUTINE: shFree
    *
    * Attempt to free the given piece of memory.  
    * 
    * RETURNS:
    *      nothing
    */

void 
shFree
   (
   void *vptr                /* I: free this chunk of memory */
   )
{
   free(vptr);
}



   /*********************************************************************
    * ROUTINE: shError
    *
    * Print the given error message to stderr, but continue to execute.
    * 
    * RETURNS:
    *      nothing
    */

void 
shError
   (
   char *format,             /* I: format part of printf statement */
   ...                       /* I: optional arguments to printf */
   )
{
   va_list ap;

   va_start(ap, format);
   (void) vfprintf(stderr, (const char *)format, ap);
   fputc('\n', stderr);
   fflush(stdout);
   fflush(stderr);
   va_end(ap);
}


   /*********************************************************************
    * ROUTINE: shFatal
    *
    * Print the given error message to stderr, and halt program execution.
    * 
    * RETURNS:
    *      nothing
    */

void 
shFatal
   (
   char *format,             /* I: format part of printf statement */
   ...                       /* I: optional arguments to printf */
   )
{
   va_list ap;

   va_start(ap, format);
   (void) vfprintf(stderr, (const char *)format, ap);
   fputc('\n', stderr);
   fflush(stdout);
   fflush(stderr);
   va_end(ap);
   exit(1);
}


   /*********************************************************************
    * ROUTINE: shDebugSet, shDebug
    *
    * shDebugSet: sets the static variable 'debug_level' to the
    * given integer; it starts at 0, but the user may set it to
    * higher levels.  The higher the level, the more messages
    * may be printed during execution.
    *
    * shDebug: If the current 'debug level' is >= the passed 'level', 
    * then print the given message to stdout, and continue execution.
    * Otherwise, just continue.
    * 
    * RETURNS:
    *      nothing
    */

static int debug_level = 0;

void
shDebugSet
   (
   int level                 /* I: set debug level to this value */
   )
{
   debug_level = level;
}

void 
shDebug
   (
   int level,                /* I: debug level at which we print this */
   char *format,             /* I: format part of printf statement */
   ...                       /* I: optional arguments to printf */
   )
{
   va_list ap;

   if (level > debug_level) {
      return;
   }

   va_start(ap, format);
   (void) vfprintf(stdout, (const char *)format, ap);
   fputc('\n', stdout);
   fflush(stdout);
   va_end(ap);
}



/************************************************************************
 * 
 *
 * ROUTINE: obsNew
 *
 * DESCRIPTION:
 * Create a new OBS structure, and return a pointer to it.
 *
 * RETURN:
 *    TRANS *           if all goes well
 *    NULL              if not
 *
 * </AUTO>
 */

OBS *
obsNew
   (
	int id,
	double time,
	double x,
	double y,
	double z
   )
{
   OBS *new;

   new = shMalloc(sizeof(OBS));
   new->id = id;
	new->time = time;
	new->x = x;
	new->y = y;
	new->z = z;
	new->next = (OBS *) NULL;
   return(new);
}

/************************************************************************
 * 
 *
 * ROUTINE: obsDel
 *
 * DESCRIPTION:
 * Delete the given OBS structure
 *
 * RETURN:
 *    nothing
 *
 * </AUTO>
 */

void
obsDel
   (
   OBS *obs                 /* I: structure to delete */
   )
{
   shFree(obs);
}



/************************************************************************
 * 
 * ROUTINE: copy_obs_list
 *
 * DESCRIPTION:
 * Make a copy of the given list of OBS structures,
 *     and return a pointer to the head of the new list.
 *
 * RETURN:
 *    a pointer to the head of the new copy
 *
 * </AUTO>
 */

OBS *
copy_obs_list
   (
   OBS *orig_list                 /* I: head of original list */
   )
{
	OBS *old_obs, *new_obs, *new_head, *new_prev;

	shAssert(orig_list != NULL);

	new_head = (OBS *) NULL;
	
	old_obs = orig_list;
	while (old_obs != NULL) {
		new_obs = obsNew(old_obs->id, old_obs->time, 	
								old_obs->x, old_obs->y, old_obs->z);

		old_obs = old_obs->next;

		/* must be careful here to take care of first case */
		if (new_head == NULL) {
			new_head = new_obs;
			new_prev = new_obs;
		}
		else {
			new_prev->next = new_obs;
			new_prev = new_obs;
		}
	}

	return(new_head);
}



/************************************************************************
 * 
 * ROUTINE: compare_obs_lists
 *
 * DESCRIPTION:
 * Given two lists of OBS, walk through them and compare the
 *     properties of each element to the corresponding element
 *     in the other list.
 *
 * RETURN:
 *     0       if every element matches its partner 
 *     N       the number of differences between the lists
 *
 * </AUTO>
 */

int
compare_obs_lists
   (
   OBS *list_A,        /* I: compare elements of this list ... */
   OBS *list_B         /* I: .. to elements of this list ... */
   )
{
	int num_fail;
	OBS *obs_A, *obs_B;

	shAssert(list_A != NULL);
	shAssert(list_B != NULL);
	num_fail = 0;

	obs_A = list_A;
	obs_B = list_B;
	while (obs_A != NULL) {
		if (obs_A->id != obs_B->id) {
			num_fail++;
		}
		if (obs_A->time != obs_B->time) {
			num_fail++;
		}
		if (obs_A->x != obs_B->x) {
			num_fail++;
		}
		if (obs_A->y != obs_B->y) {
			num_fail++;
		}
		if (obs_A->z != obs_B->z) {
			num_fail++;
		}
	
		obs_A = obs_A->next;
		obs_B = obs_B->next;
	}
	
	/* careful!  If list B has extra elements, that's failure, too */
	if (obs_B != NULL) {
		num_fail++;
	}
	if (verbose) {
		printf("compare_obs_list finds %d differences \n", num_fail);
	}

	return(num_fail);
}






/*********************************************************************
 * ROUTINE: read_obs_file
 *
 * Given the name of a file, and four integers which specify the
 * columns in which the "time", "x", "y", and "z" data appear, go through
 * the file, creating an "s_obs" structure for each line.
 * Place all the objs in a linked list, and return a pointer
 * to the head of the list, and the number of stars in the list.
 *
 * Ignore any line which starts with a COMMENT_CHAR, and any
 * completely empty line (one with whitespace only).
 *
 * Note that columns are numbered starting at 0.  Thus, 
 * given a file with format like this:
 *   
 *    #    x       y      mag
 *       234.43  54.33   12.23
 *
 * we have xcolumn = 0, ycolumn = 1, magcolumn = 2.
 *
 * RETURNS:
 *      SH_SUCCESS         if all goes well
 *      SH_GENERIC_ERROR   if not
 */

int
read_obs_file
   (
   char *filename,           /* I: name of file */
   int timecolumn,           /* I: column in which 'time' data is found */
   int xcolumn,              /* I: column in which 'x' data is found */
   int ycolumn,              /* I: column in which 'y' data is found */
   int zcolumn,              /* I: column in which 'z' data is found */
   int *num_obs,             /* O: number of obs in new list goes here */
   struct s_obs **list       /* O: new OBS list will be placed here */
   )
{
   FILE *fp;
   char line[LINELEN];
   char col[MAX_DATA_COL + 1][MAX_COL_LENGTH + 1];
   int nline, num, ncol;
   int last_column = -1;
   struct s_obs *head, *last, *new;
   double timeval, xval, yval, zval;

   /* sanity checks */
   shAssert(timecolumn >= 0);
   shAssert((xcolumn >= 0) && (xcolumn != timecolumn));
   shAssert((ycolumn >= 0) && (ycolumn != timecolumn) &&
                              (ycolumn != xcolumn));
   shAssert((zcolumn >= 0) && (zcolumn != timecolumn) &&
            (zcolumn != xcolumn) && (zcolumn != ycolumn));

   if ((fp = fopen(filename, "r")) == NULL) {
      shError("read_obs_file: can't open file %s\n", filename);
      return(SH_GENERIC_ERROR);
   }

   /* find the highest-numbered column we need to read */
   last_column = timecolumn;
   if (xcolumn > last_column) {
     last_column = xcolumn;
   }
   if (ycolumn > last_column) {
     last_column = ycolumn;
   }
   if (zcolumn > last_column) {
     last_column = zcolumn;
   }
   shAssert(last_column >= 3);
   if (last_column > MAX_DATA_COL) {
      shError("read_obs_file: only %d columns allowed", MAX_DATA_COL);
      return(SH_GENERIC_ERROR);
   }

   nline = 0;
   head = (OBS *) NULL;
   last = head;
   num = 0;
   while (fgets(line, LINELEN, fp) != NULL) {
      if (line[0] == COMMENT_CHAR) {
         nline++;
         continue;
      }
      if (is_blank(line)) {
         nline++;
         continue;
      }
      ncol = sscanf(line, "%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s", 
              &(col[0][0]), &(col[1][0]), &(col[2][0]), &(col[3][0]), 
              &(col[4][0]), &(col[5][0]), &(col[6][0]), &(col[7][0]), 
              &(col[8][0]), &(col[9][0]),
				  &(col[10][0]), &(col[11][0]), &(col[12][0]),
				  &(col[13][0]), &(col[14][0]), &(col[15][0]),
				  &(col[16][0]), &(col[17][0]), &(col[18][0]),
				  &(col[19][0]),
				  &(col[20][0]), &(col[21][0]), &(col[22][0]),
				  &(col[23][0]), &(col[24][0]), &(col[25][0]),
				  &(col[26][0]), &(col[27][0]), &(col[28][0]),
				  &(col[29][0]));
      if (last_column > ncol) {
         shError("read_obs_file: not enough entries in following line; skipping");
         shError("  %s", line);
         nline++;
         continue;
      }
        
      /* now read values from each column */
      if (get_value(col[timecolumn], &timeval) != SH_SUCCESS) {
         shError("read_obs_file: can't read time value from %s; skipping", 
                  col[timecolumn]);
         nline++;
         continue;
      }
      if (get_value(col[xcolumn], &xval) != SH_SUCCESS) {
         shError("read_obs_file: can't read X value from %s; skipping", 
                  col[xcolumn]);
         nline++;
         continue;
      }
      if (get_value(col[ycolumn], &yval) != SH_SUCCESS) {
         shError("read_obs_file: can't read Y value from %s; skipping", 
                  col[ycolumn]);
         nline++;
         continue;
      }
      if (get_value(col[zcolumn], &zval) != SH_SUCCESS) {
         shError("read_obs_file: can't read z value from %s; skipping", 
                  col[zcolumn]);
         nline++;
         continue;
      }


      /* okay, it's safe to create a new s_obs for this line */
      new = obsNew(num, timeval, xval, yval, zval);
      if (head == NULL) {
         head = new;
         last = new;
      }
      else {
         last->next = new;
         last = new;
      }

      nline++;
      num++;
   }

   *num_obs = num;
   *list = head;

   return(SH_SUCCESS);
}



/**********************************************************************
 * ROUTINE: is_blank
 *
 * If the given string consists only of whitespace, return 1.
 * Otherwise, return 0.
 */

int 
is_blank
   (
   char *line                /* I: string to be checked */
   )
{
   char *p;

   for (p = line; (*p != '\0') && (isspace(*p)); p++) {
      ;
   }
   /* 17/Dec/01, jpb: fixed a bug that said "if (p=='\0')" */
   if (*p == '\0') {
      return(1);
   }
   else {
      return(0);
   }
}


/**********************************************************************
 * ROUTINE: get_value
 *
 * Given a string containing a numerical value, read the numerical
 * value and place it into the given double argument.  
 *
 * Return 0 if all goes well.
 * Return 1 if there is an error.
 */

int 
get_value
   (
   char *str,                /* I: string to be converted to double */
   double *val               /* O: place value here */
   )
{
   if (sscanf(str, "%lf", val) != 1) {
      return(1);
   } 
   else {
      return(0);
   }
}
 


/**********************************************************************
 * ROUTINE: print_obs
 *
 * Given a pointer to an OBS structure, print out its contents
 * in a single line.
 *
 * RETURNS: void
 */

void
print_obs
   (
	OBS *obs
   )
{
   shAssert(obs != NULL);
	
	printf("  %12d %12.5e  %12.4e %12.4e %12.4e \n",
		obs->id, obs->time, obs->x, obs->y, obs->z);

}
 

/**********************************************************************
 * ROUTINE: print_obs_list
 *
 * Given a pointer to a list of OBS structure, print out the structures
 *    in the list: one line per structure.
 *
 * RETURNS:  void
 */

void 
print_obs_list
   (
	OBS *obs
   )
{
   shAssert(obs != NULL);
	
	print_obs(obs);
	while (obs->next != NULL) {
		obs = obs->next;
		print_obs(obs);
	}
}
 

