
	/*****************************************************
	 * Perform a series of calculations to analyze a set
	 *    of "observations" of the position of a host star
	 *    over time, the goal being to determine the properties
	 *    of a single planet orbiting the star.
	 *
	 * We ignore all observational uncertainties.  We assume
	 *    that parallax and proper motion have already been removed.
	 *
	 * MWR 10/2/2016
	 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "misc.h"


	/* these are the default (0-index) columns of data in the input file */
	 
#define TIMECOL  0
#define XCOL     3
#define YCOL     4
#define ZCOL     5


	/*
	 * USAGE: solve_ellipse  infile=   [verbose]  [help] 
	 */
#define USAGE "solve_ellipse infile=  [verbose]  [help]  "

#define DEBUG


	/* global variables.  Yeah, whatever */
#ifdef DEBUG
int verbose = 1;
#else
int verbose = 0;
#endif

int
main
	(
	int argc,
	char *argv[]
	)
{
	int i;
	char infile[LINELEN];

	if ((argc < 2) ||
	    (strncmp(argv[1], "help", 4) == 0)) {
		printf("%s\n", USAGE);
		exit(0);
	}


	/* parse args */
	for (i = 1; i < argc; i++) {
		if (verbose > 0) {
			printf(" arg %4d is ..%s.. \n", i, argv[i]);
		}
		if (strncmp(argv[i], "infile=", 7) == 0) {
			if (sscanf(argv[i] + 7, "%s", infile) != 1) {
				shFatal("can't parse infile arg %s \n", argv[i]);
			}
		}
		else if (strncmp(argv[i], "verbose", 7) == 0) {
			verbose = 1;
		}
		else {
			/* this is not a recognized argument.  Complain and quit */
			shFatal("invalid argument: %s", argv[i]);
		} 
	}
				

	/* 
	 * step 1: read in the "observations", which are really the
	 *         perfect results of an N-body simulation.
	 *         We read in position of the HOST STAR at a number of times
	 *         over one complete revolution around the center of mass.
	 *  
	 *         We do NOT read in any information on the planet.
	 */
	
	/* number of 'observations' of the host star */
	int num_obs;   
	/* head of the linked list of all observations */
	OBS *orig_list, *center_list;
	
	if (read_obs_file(infile, TIMECOL, XCOL, YCOL, ZCOL, &num_obs, &orig_list)
				!= 0) {
		shFatal("read_obs_file fails for file %s \n", infile);
	}
	if (verbose) {
		printf(" read %d obs from %s \n", num_obs, infile);
		print_obs_list(orig_list);
	}

	/* 
	 * create a copy of the original data, 'center_obs', and 
	 *     re-center it so that it sits at the origin
	 */
	if ((center_list = copy_obs_list(orig_list)) == NULL) {
		shFatal("copy_obs_list fails before centering");
	}
		/* this is a sanity check */
		if (compare_obs_lists(orig_list, center_list) != 0) {
			shFatal("compare_obs_lists fails for center_list");
		}
	if (recenter_ellipse(center_list) != 0) {
		shFatal("recenter_ellipse fails \n");
	}
	if (verbose > 0) {
		printf("# here comes recentered list \n");
		print_obs_list(center_list);
	}
	
	
	
	




	exit(0);
}


	 
